
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <GL/glew.h>
#include <GL/glut.h>
#include <vector>

#include "asteroids.h"
#include "collision.h"
#include "shaders.h"
#include "sdlbase.h"
#include "vec.h"
#include "effects.h"

using namespace std;

#define DEFAULT_FOV 75 /* perspective projection field of view */
#define DEFAULT_NEAR 0.1 /* near clipping plane */
#define DEFAULT_FAR 100.0 /* far clipping plane */

#define RAND_UNIT (rand()/(float)RAND_MAX)
static float pi;
static vec2f totalMomentum;
static float totalKineticEnergy = 0;
float tickCounter = 0;
static vec2f edgeMomentum;
static bool squareBound;
/* global debug flags */
typedef enum {
	DEBUG_SHIP,
	DEBUG_KINETIC,
	DEBUG_MOMENTUM,
	DEBUG_COLLIDE,
	DEBUG_FLAG_COUNT
} DebugFlag;
bool debugFlag[DEBUG_FLAG_COUNT];

/* global game variables */
typedef enum {
	GAMESTATE_GAMEOVER,
	GAMESTATE_WIN,
	GAMESTATE_INGAME
} GameState;
static GameState gameState = GAMESTATE_GAMEOVER;

/* asteroids variables */
const int initNumAsteroids = 10;
const int numAsteroidSplits = 3;
const float initAsteroidSpeed = 0.5;
const float splitForce = 0.035;

static vector<asteroid_t> asteroids;

/* NOTE: use "asteroids.erase(asteroids.begin() + i);" to remove an element */

static vec2f levelBounds;
static vec2f centre;

/* ship */
static ship_t ship;

/* missiles */
#define NUM_MISSILES 5
const float MISSILE_SPEED = 1.0;
static vector<missile_t> missiles;
Explosion explosion;
Star stars;
float displacement[100];

void initDisplacement(){
	int i;
	for(i = 0; i < 100; i++){
		displacement[i] = RAND_UNIT/50;
	}
}

void splitAsteroid(int i){
	asteroid_t first, second;
	explosion.explode(asteroids[i].p);
	
	asteroids[i].life -= 1;
	if(asteroids[i].life == 0){
		asteroids.erase(asteroids.begin() + i);
		return;
	}
	first.life = asteroids[i].life;
	second.life = asteroids[i].life;
	
	first.r = asteroids[i].r/2;
	second.r = first.r;
	
	first.v = asteroids[i].v;
	second.v = -first.v;
	
	vec2f dir = second.v;
	dir.normalize();
	
	first.p = asteroids[i].p + (dir * first.r);
	
	second.p.x = first.p.x - (first.r*2 * dir.x);
	second.p.y = first.p.y - (first.r*2 * dir.y);
	
	first.rotation = 0;
	first.rotationRate = (RAND_UNIT-0.5)/100;
	first.randomDisplacement = floor(RAND_UNIT * 50);
	second.rotation = 0;
	second.rotationRate = (RAND_UNIT-0.5)/100;
	second.randomDisplacement = floor(RAND_UNIT * 60);
	
	asteroids.erase(asteroids.begin() + i);
	asteroids.push_back(first);
	asteroids.push_back(second);
}

void initAsteroids(int count)
{
	bool good = false;
	/* create asteroids */
	//asteroids.resize(count);
	
	/* initialize asteroids */
	for (int i = 0; i < initNumAsteroids; ++i)
	{
		good = false;
		asteroid_t newAsteroid;
		newAsteroid.r = 0.1;
		newAsteroid.p.x = RAND_UNIT * (levelBounds.x - 2.0 * newAsteroid.r) + newAsteroid.r;
		newAsteroid.p.y = RAND_UNIT * (levelBounds.y - 2.0 * newAsteroid.r) + newAsteroid.r;

		while(good == false){
			while(newAsteroid.p.x < 2.5 && newAsteroid.p.x > 1.5 && newAsteroid.p.y < 2 && newAsteroid.p.y > 1){
				newAsteroid.p.x = RAND_UNIT * (levelBounds.x - 2.0 * newAsteroid.r) + newAsteroid.r;
				newAsteroid.p.y = RAND_UNIT * (levelBounds.y - 2.0 * newAsteroid.r) + newAsteroid.r;
			}
			good = true;
			for(int j = 0; j < (int)asteroids.size(); j++){
				if(intersectCircleCircle(asteroids[j].p, asteroids[j].r, newAsteroid.p, newAsteroid.r)) {
					newAsteroid.p.x = RAND_UNIT * (levelBounds.x - 2.0 * newAsteroid.r) + newAsteroid.r;
					newAsteroid.p.y = RAND_UNIT * (levelBounds.y - 2.0 * newAsteroid.r) + newAsteroid.r;
					good = false;
					break;
				}
			}
		}
		
		newAsteroid.v.x = RAND_UNIT * 2.0 - 1.0;
		newAsteroid.v.y = RAND_UNIT * 2.0 - 1.0;
		newAsteroid.v = newAsteroid.v.unit() * initAsteroidSpeed;
		
		newAsteroid.rotation = 0;
		newAsteroid.rotationRate = (RAND_UNIT-0.5)/100;
		newAsteroid.randomDisplacement = floor(RAND_UNIT * 60);
		newAsteroid.life = numAsteroidSplits;
		asteroids.push_back(newAsteroid);
	}
}

void fireMissile()
{
	/* don't fire too many */
	if (missiles.size() >= NUM_MISSILES)
		return;

	missile_t newMissile;
	vec2f dir = vec2f(-sin(ship.heading), cos(ship.heading));
	newMissile.p = ship.p + dir * ship.size;
	newMissile.v = ship.v + dir * MISSILE_SPEED;
	newMissile.r = 0.01;
	newMissile.life = 2.0;
	missiles.push_back(newMissile);
}

void reshape(int w, int h)
{
	float aspect;
	if (h == 0) h = 1; /* stops division by zero */
	glViewport(0, 0, w, h); /* draw to the entire window */
	aspect = w / (float)h; /* calculate aspect ratio */
	glMatrixMode(GL_PROJECTION); /* we are now modifying the projection matrix */
	glLoadIdentity(); /* clear previous values */

	/* set a perspective projection */
	glOrtho(0.0, levelBounds.x, 0.0, levelBounds.y, -1.0, 1.0);

	glMatrixMode(GL_MODELVIEW); /* don't forget to go back to the modelview matrix */
}

void restart()
{
	initAsteroids(initNumAsteroids);

	ship.p.x = levelBounds.x * 0.5;
	ship.p.y = levelBounds.y * 0.5;
	ship.v.x = 0.0;
	ship.v.y = 0.0;
	ship.heading = 0.0;
	ship.thruster = false;
	ship.turnLeft = false;
	ship.turnRight = false;
	ship.rThruster = false;
	ship.engine.erase();
	ship.explosion.erase();

	gameState = GAMESTATE_INGAME;

	stars.erase();
	stars.setSize(levelBounds);
	stars.createStars();
	missiles.clear();
	explosion.erase();
}

void init()
{
	pi = acos(-1.0);
	levelBounds.x = 4.0;
	levelBounds.y = 3.0;
	edgeMomentum = 0;
	ship.size = 0.1;
	ship.w = 2.0;
	ship.a = 1.0;
	ship.lives = 3;
	ship.lastHit = 0.0;
	squareBound = true;
	
	initDisplacement();
	memset(debugFlag, 0, sizeof(bool) * DEBUG_FLAG_COUNT);

	restart();
}

void drawRegularPolygon(float radius, int subs, bool displace, int displaceSeed)
{
	/* draws a polygon with same length sides */
	int i;
	float theta;
	float Dradius;
	
	glPushAttrib(GL_CURRENT_BIT);
	glBegin(GL_LINE_LOOP);
	for (i = 0; i <= subs; ++i) {
		theta = (2.0 * i * pi / subs);
		if(displace){
			Dradius = radius + displacement[i + displaceSeed];
			#ifdef COLOUR
			glColor3f(0.5,0.2,0);
			#endif
		}
		else{
			Dradius = radius;
		}
		glVertex2f(Dradius * cos(theta), Dradius * sin(theta));
	}
	glEnd();
	glPopAttrib();
}

void drawAxes(float length)
{
	glPushAttrib(GL_CURRENT_BIT);
	glPushMatrix();
	glScalef(length, length, length);
	glBegin(GL_LINES);
	glColor3f(1.0, 0.0, 0.0);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(1.0, 0.0, 0.0);
	glColor3f(0.0, 1.0, 0.0);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(0.0, 1.0, 0.0);
	glEnd();
	glPopMatrix();
	glPopAttrib();
}

void displayHUD(){
	char score[15];
	char *bufp;
	/* draw the ship */
	glPushAttrib(GL_CURRENT_BIT);
	glPushMatrix();
#ifdef COLOUR
	glColor3f(0, 0.5, 0);
#endif
	snprintf(score, sizeof(score), "Score: %d", ship.score);
	glRasterPos2f(0.1,levelBounds.y-0.15);
	for(bufp = score; *bufp; bufp++)
            glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *bufp);
	
	if(ship.lives >= 1){
		glTranslatef(0.15, levelBounds.y-0.5, 0.0);
		glBegin(GL_LINE_LOOP);
		glVertex2f(0.0, 0.1 * 2.0);
		glVertex2f(-0.05, 0.0);
		glVertex2f(0.05, 0.0);
		glEnd();
	}
	if(ship.lives >= 2){
		glTranslatef(0.15, 0.0, 0.0);
		glBegin(GL_LINE_LOOP);
		glVertex2f(0.0, 0.1 * 2.0);
		glVertex2f(-0.05, 0.0);
		glVertex2f(0.05, 0.0);
		glEnd();
	}
	if(ship.lives >= 3){
		glTranslatef(0.15, 0.0, 0.0);
		glBegin(GL_LINE_LOOP);
		glVertex2f(0.0, 0.1 * 2.0);
		glVertex2f(-0.05, 0.0);
		glVertex2f(0.05, 0.0);
		glEnd();
	}
	glPopMatrix();
	glPopAttrib();
}

void showWinMessage()
{
	char win[64];
	char *bufp;
	/* draw the ship */
	glPushAttrib(GL_CURRENT_BIT);
	glPushMatrix();
#ifdef COLOUR
	glColor3f(0, 0.5, 0);
#endif
	snprintf(win, sizeof(win), "Congratulations! You have won the game!");
	glRasterPos2f(1,1.6);
	for(bufp = win; *bufp; bufp++)
            glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *bufp);
    
    snprintf(win, sizeof(win), "Your score was %d",ship.score);
	glRasterPos2f(1.5,1.4);
	for(bufp = win; *bufp; bufp++)
            glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *bufp);
	
	glPopMatrix();
	glPopAttrib();
}

void showGameOverMessage()
{
	char lose[64];
	char *bufp;
	/* draw the ship */
	glPushAttrib(GL_CURRENT_BIT);
	glPushMatrix();
#ifdef COLOUR
	glColor3f(0, 0.5, 0);
#endif
	snprintf(lose, sizeof(lose), "GAME OVER - YOU ARE OUT OF LIVES");
	glRasterPos2f(1.2,1.6);
	for(bufp = lose; *bufp; bufp++)
            glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *bufp);
    
    snprintf(lose, sizeof(lose), "Your score was %d",ship.score);
	glRasterPos2f(1.5,1.4);
	for(bufp = lose; *bufp; bufp++)
            glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *bufp);
	
	glPopMatrix();
	glPopAttrib();
}

void display()
{
	int i;
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	
	if(gameState == GAMESTATE_WIN)
		showWinMessage();
	else if(gameState == GAMESTATE_GAMEOVER && ship.lives == 0)
		showGameOverMessage();
	
	stars.drawPoints();
	
	glPushMatrix();
	glLineWidth(3);
	glColor3f(1,0,0);
	if(!squareBound) {
		glTranslatef(2,1.5,0);
		drawRegularPolygon(1.5,64,false,0);
	} else {
		glBegin(GL_LINE_LOOP);
		glVertex3f(0.01,0.01,0);
		glVertex3f(3.99,0.01,0);
		glVertex3f(3.99,2.99,0);
		glVertex3f(0.01,2.99,0);
		glEnd();
	}
	glLineWidth(1);
	glPopMatrix();
	
	
	/* draw the asteroids */
	glColor3f(1.0, 1.0, 1.0);
	for (i = 0; i < (int)asteroids.size(); ++i)
	{
		glPushMatrix();
		glTranslatef(asteroids[i].p.x, asteroids[i].p.y, 0.0);
		glRotatef(asteroids[i].rotation,0,0,1);
		drawRegularPolygon(asteroids[i].r, 32, true, asteroids[i].randomDisplacement);
		glPopMatrix();
	}

	/* draw the missiles */
	for (i = 0; i < (int)missiles.size(); ++i)
	{
		glPushMatrix();
		glColor3f(0.25,0.25,1);
		glTranslatef(missiles[i].p.x, missiles[i].p.y, 0.0);
		drawRegularPolygon(missiles[i].r, 6, false, 0);
		missiles[i].plasma.drawPoints();
		glPopMatrix();
	}

	/* draw the ship */
	glPushMatrix();
	glTranslatef(ship.p.x, ship.p.y, 0.0);
	if (debugFlag[DEBUG_SHIP])
		drawAxes(0.5);
	glRotatef(ship.heading * 180.0/pi, 0.0, 0.0, 1.0);
	if (debugFlag[DEBUG_SHIP])
		drawAxes(0.5);
	glTranslatef(0.0, -ship.size, 0.0);
	glBegin(GL_LINE_LOOP);
#ifdef COLOUR
	glColor3f(0, 1, 0);
#endif
	glVertex2f(0.0, ship.size * 2.0);
	glVertex2f(-0.5 * ship.size, 0.0);
	glVertex2f(0.5 * ship.size, 0.0);
	glEnd();
	glPopMatrix();
	
	ship.engine.drawPoints();
	
	explosion.drawPoints();
	displayHUD();
	ship.explosion.drawPoints();
	
	checkError();
}

void updateAsteroid(int i, float dt)
{
	/* movement integration */
	asteroids[i].p += asteroids[i].v * dt;
	float force;
	vec2f momInit = asteroids[i].v * asteroids[i].r;
	if(debugFlag[DEBUG_MOMENTUM])
		totalMomentum += momInit;
	if(debugFlag[DEBUG_KINETIC])
		totalKineticEnergy += 0.5 * asteroids[i].r * asteroids[i].v.sizesq();
	
	if(squareBound) {
		/* level bounds collision */
		if (intersectCircleHorLine(asteroids[i].p, asteroids[i].r, 0.0)){
			force = reactCircleHorLine(&asteroids[i].p, &asteroids[i].v, asteroids[i].r, 0.0);
			if(force > splitForce){
				splitAsteroid(i);
			}
		}
		if (intersectCircleHorLine(asteroids[i].p, asteroids[i].r, levelBounds.x)){
			force = reactCircleHorLine(&asteroids[i].p, &asteroids[i].v, asteroids[i].r, levelBounds.x);
			if(force > splitForce){
				splitAsteroid(i);
			}
		}
		if (intersectCircleVerLine(asteroids[i].p, asteroids[i].r, 0.0)){
			force = reactCircleVerLine(&asteroids[i].p, &asteroids[i].v, asteroids[i].r, 0.0);
			if(force > splitForce){
				splitAsteroid(i);
			}
		}
		if (intersectCircleVerLine(asteroids[i].p, asteroids[i].r, levelBounds.y)){
			force = reactCircleVerLine(&asteroids[i].p, &asteroids[i].v, asteroids[i].r, levelBounds.y);
			if(force > splitForce){
				splitAsteroid(i);
			}
		}
	} else {
		if(intersect(asteroids[i],levelBounds))
			react(asteroids[i],levelBounds);
	}
	vec2f momFinal = asteroids[i].v * asteroids[i].r;
	edgeMomentum += momFinal - momInit;
	/* collide asteroid with ship */
	if (intersectCircleCircle(asteroids[i].p, asteroids[i].r, ship.p, ship.size)) {
		gameState = GAMESTATE_GAMEOVER;
		ship.lives --;
	}
	for(int j = i+1; j < (int)asteroids.size(); j++) {
		if(intersectCircleCircle(asteroids[i].p,asteroids[i].r,asteroids[j].p,asteroids[j].r)) {
			force = reactAsteroids(asteroids[i],asteroids[j],debugFlag[DEBUG_COLLIDE]);
			if(force > splitForce){
				splitAsteroid(i);
				splitAsteroid(j);
			}
		}
	}
	asteroids[i].rotation += asteroids[i].rotationRate*dt*10000;
}


void updateMissile(int i, float dt)
{
	/* movement integration */
	missiles[i].p += missiles[i].v * dt;

	/* level bounds collision */
	if (intersectCircleHorLine(missiles[i].p, missiles[i].r, 0.0)){
		missiles.erase(missiles.begin() + i);
		return;
	}
	else if (intersectCircleHorLine(missiles[i].p, missiles[i].r, levelBounds.x)){
		missiles.erase(missiles.begin() + i);
		return;
	}
	else if (intersectCircleVerLine(missiles[i].p, missiles[i].r, 0.0)){
		missiles.erase(missiles.begin() + i);
		return;
	}
	else if (intersectCircleVerLine(missiles[i].p, missiles[i].r, levelBounds.y)){
		missiles.erase(missiles.begin() + i);
		return;
	}
	
	for(int j = 0; j < (int)asteroids.size(); j++){
		if(intersectCircleCircle(missiles[i].p, missiles[i].r, asteroids[j].p, asteroids[j].r)){
			if(ship.lastHit == asteroids[j].r){
				ship.score += 5;
			}
			ship.lastHit = asteroids[j].r;
			ship.score += 10;
			missiles.erase(missiles.begin() + i);
			splitAsteroid(j);
			return;
		}
	}
	
	missiles[i].life -= dt;
	if(missiles[i].life <= 0){
		missiles.erase(missiles.begin() + i);
		return;
	}
	missiles[i].plasma.updatePoints(dt);
	missiles[i].plasma.createPoint(missiles[i].p, 0);
	missiles[i].plasma.createPoint(missiles[i].p, 0);
}

void update(float dt)
{
	int i;
	if(asteroids.size() == 0)
		gameState = GAMESTATE_WIN;
		
	if(debugFlag[DEBUG_MOMENTUM])
		totalMomentum = 0;
	if(debugFlag[DEBUG_KINETIC])
		totalKineticEnergy = 0;
		
	if (gameState != GAMESTATE_INGAME){
		if(gameState == GAMESTATE_GAMEOVER) {
			ship.explosion.explode(ship.p);
			ship.explosion.updatePoints(dt);
		}
		return;
	}

	/* asteroids physics */
	for (i = (int)asteroids.size(); i > 0; --i){
		/* Note: Iteration is done this way to handle dynamic size decreases of 1 */
		updateAsteroid((int)asteroids.size() - i, dt);
		//printf("%d\n", (int)asteroids.size() - i);
	}

	/* missile physics */
	for (i = (int)missiles.size(); i > 0; --i)
		updateMissile((int)missiles.size() - i, dt);

	/* ship physics */
	if (ship.turnLeft)
		ship.heading += ship.w * dt;
	if (ship.turnRight)
		ship.heading -= ship.w * dt;
	vec2f shipAcc;
	if (ship.thruster){
		shipAcc = vec2f(-sin(ship.heading), cos(ship.heading)) * (ship.a * dt);
		ship.engine.createPoint(ship.p, ship.heading);
	}
	else if (ship.rThruster){
		shipAcc = -vec2f(-sin(ship.heading), cos(ship.heading)) * (ship.a * dt);
	}
	else
		shipAcc = vec2f(0.0, 0.0);
	/* midpoint numerical integration */
	ship.p += (ship.v + shipAcc * 0.5) * dt;
	ship.v += shipAcc;

	/* ship/level bounds collision */
	if(squareBound) {
		if (intersectCircleHorLine(ship.p, ship.size, 0.0))
			reactCircleHorLine(&ship.p, &ship.v, ship.size, 0.0);
		if (intersectCircleHorLine(ship.p, ship.size, levelBounds.x))
			reactCircleHorLine(&ship.p, &ship.v, ship.size, levelBounds.x);
		if (intersectCircleVerLine(ship.p, ship.size, 0.0))
			reactCircleVerLine(&ship.p, &ship.v, ship.size, 0.0);
		if (intersectCircleVerLine(ship.p, ship.size, levelBounds.y))
			reactCircleVerLine(&ship.p, &ship.v, ship.size, levelBounds.y);
	} else {
		if(intersect(ship,levelBounds))
			react(ship,levelBounds);
	}
		
	ship.engine.updatePoints(dt);
	explosion.updatePoints(dt);
	
	tickCounter += dt;
	if(debugFlag[DEBUG_MOMENTUM])
		totalMomentum -= edgeMomentum;
	if(tickCounter >= 1.0) {
		tickCounter -= 1.0;
		if(debugFlag[DEBUG_KINETIC])
			printf("Kinetic Energy: %f\n",totalKineticEnergy);
		if(debugFlag[DEBUG_MOMENTUM])
			printf("Total Momentum: %f\n",totalMomentum.size());
	}
}

void mouseDown(int button, int state, int x, int y)
{
}

void mouseMove(int x, int y)
{
}

void keyDown(int key)
{
	if (key == SDLK_ESCAPE)
		quit();
	if (gameState != GAMESTATE_INGAME) {
		if (key == SDLK_SPACE){
			if(ship.lives == 0){
				ship.lives = 3;
				ship.score = 0;
			}
			asteroids.erase(asteroids.begin(), asteroids.end());
			restart();
		}
		return;
	}

	/* ship controls */
	if (key == SDLK_UP)
		ship.thruster = true;
	if (key == SDLK_DOWN)
		ship.rThruster = true;
		
	if (key == SDLK_LEFT)
		ship.turnLeft = true;
	if (key == SDLK_RIGHT)
		ship.turnRight = true;
	if (key == SDLK_SPACE)
		fireMissile();

	/* toggles entire debugFlag array. for better control, add unique toggles */
	if (key == SDLK_d)
		memset(debugFlag, !debugFlag[0], sizeof(bool) * DEBUG_FLAG_COUNT);
	if (key == SDLK_f)
		debugFlag[DEBUG_COLLIDE] = false;
	if (key == SDLK_c)
		squareBound = !squareBound;
}

void keyUp(int key)
{
	if (key == SDLK_UP)
		ship.thruster = false;
	if (key == SDLK_DOWN)
		ship.rThruster = false;
	if (key == SDLK_LEFT)
		ship.turnLeft = false;
	if (key == SDLK_RIGHT)
		ship.turnRight = false;
}

void cleanup()
{
}

